from functools import partial
from pathlib import Path

from poetry.console.commands.command import Command
from polylith import commands, configuration, info, repo
from polylith.poetry.commands import command_options
from polylith.poetry.internals import (
    distributions,
    filter_projects_data,
    merge_project_data,
)


class LibsCommand(Command):
    name = "poly libs"
    description = "Show third-party libraries used in the workspace."

    options = [
        command_options.alias,
        command_options.save,
        command_options.short,
        command_options.strict,
    ]

    def merged_project_data(self, project_data: dict) -> dict:
        name = project_data["name"]

        try:
            return merge_project_data(project_data)
        except ValueError as e:
            self.line_error(f"{name}: <error>{e}</error>")
            return project_data

    def handle(self) -> int:
        root = repo.get_workspace_root(Path.cwd())
        dists_fn = partial(distributions, root)
        save = self.option("save")

        output = configuration.get_output_dir(root, "libs") if save else None

        options = {
            "strict": self.option("strict"),
            "alias": self.option("alias"),
            "short": self.option("short"),
            "save": save,
            "output": output,
            "dists_fn": dists_fn,
        }

        directory = self.option("directory")
        ns = configuration.get_namespace_from_config(root)

        all_projects_data = info.get_projects_data(root, ns)
        projects_data = filter_projects_data(self.poetry, directory, all_projects_data)

        merged_projects_data = [
            self.merged_project_data(data) for data in projects_data
        ]

        results = commands.libs.run(root, ns, merged_projects_data, options)
        commands.libs.run_library_versions(projects_data, all_projects_data, options)

        return 0 if all(results) else 1



from polylith.imports.grouping import (
    extract_brick_imports,
    extract_brick_imports_with_namespaces,
)
from polylith.imports.parser import (
    extract_top_ns,
    fetch_all_imports,
    fetch_excluded_imports,
    list_imports,
    parse_module,
)
from polylith.imports.usages import SYMBOLS, extract_api, fetch_brick_import_usages

__all__ = [
    "extract_brick_imports",
    "extract_brick_imports_with_namespaces",
    "extract_api",
    "extract_top_ns",
    "fetch_all_imports",
    "fetch_brick_import_usages",
    "fetch_excluded_imports",
    "list_imports",
    "parse_module",
    "SYMBOLS",
]



import ast
from functools import lru_cache
from pathlib import Path
from typing import FrozenSet, List, Set, Union

from polylith.imports import SYMBOLS, extract_api, list_imports, parse_module

PACKAGE_INTERFACE = "__init__.py"
ALL_STATEMENT = "__all__"


def target_names(t: ast.AST) -> Set[str]:
    if isinstance(t, ast.Name):
        return {t.id}

    if isinstance(t, (ast.Tuple, ast.List)):
        return {n for e in t.elts for n in target_names(e)}

    return set()


def extract_variables(statement: ast.stmt) -> Set[str]:
    if isinstance(statement, ast.Assign):
        return {n for t in statement.targets for n in target_names(t)}

    if isinstance(statement, (ast.AnnAssign, ast.AugAssign)):
        return target_names(statement.target)

    if hasattr(ast, "TypeAlias") and isinstance(statement, ast.TypeAlias):
        return {statement.name.id}

    return set()


def is_private(symbol_name: str) -> bool:
    return symbol_name.startswith("_")


@lru_cache(maxsize=None)
def parse(path: Path) -> ast.Module:
    return parse_module(path)


def extract_symbols(path: Path) -> Set[str]:
    tree = parse(path)

    return {
        s.name for s in tree.body if isinstance(s, SYMBOLS) and not is_private(s.name)
    }


def extract_public_variables(path: Path) -> Set[str]:
    tree = parse(path)

    return {v for s in tree.body for v in extract_variables(s) if not is_private(v)}


def is_the_all_statement(target: ast.expr) -> bool:
    return isinstance(target, ast.Name) and target.id == ALL_STATEMENT


def is_string_constant(expression: ast.AST) -> bool:
    return isinstance(expression, ast.Constant) and isinstance(expression.value, str)


def attribute_expr_to_parts(expr: ast.AST) -> List[str]:
    if isinstance(expr, ast.Name):
        return [expr.id]

    if isinstance(expr, ast.Attribute):
        parent = attribute_expr_to_parts(expr.value)

        return [*parent, expr.attr] if parent else []

    return []


def find_the_all_variable(statement: ast.stmt) -> Union[Set[str], None]:
    if not isinstance(statement, ast.Assign):
        return None

    if not any(is_the_all_statement(t) for t in statement.targets):
        return None

    if not isinstance(statement.value, (ast.List, ast.Tuple)):
        return None

    if not all(is_string_constant(e) for e in statement.value.elts):
        return None

    return {e.value for e in statement.value.elts if isinstance(e, ast.Constant)}


def find_the_all_pointer(statement: ast.stmt) -> Union[str, None]:
    if not isinstance(statement, ast.Assign):
        return None

    if not any(is_the_all_statement(t) for t in statement.targets):
        return None

    parts = attribute_expr_to_parts(statement.value)

    if not parts:
        return None

    *module_path, rest = parts

    if rest != ALL_STATEMENT:
        return None

    return ".".join(module_path)


def resolve_local_module_path(package_dir: Path, module_ref: str) -> Union[Path, None]:
    parts = tuple(p for p in module_ref.split(".") if p)

    if not parts:
        return None

    module_file = package_dir.joinpath(*parts).with_suffix(".py")

    if module_file.exists():
        return module_file

    module_init = package_dir.joinpath(*parts, PACKAGE_INTERFACE)

    return module_init if module_init.exists() else None


def _extract_the_all_variable(path: Path, visited: FrozenSet[Path]) -> Set[str]:
    if path in visited:
        return set()

    visited = visited | frozenset({path})

    tree = parse(path)

    literals = [find_the_all_variable(s) for s in tree.body]
    literal = next((r for r in literals if r is not None), None)

    if literal is not None:
        return literal

    pointers = (find_the_all_pointer(s) for s in tree.body)
    pointer = next((p for p in pointers if p is not None), None)

    if not pointer:
        return set()

    resolved = resolve_local_module_path(path.parent, pointer)

    return _extract_the_all_variable(resolved, visited) if resolved else set()


def extract_the_all_variable(path: Path) -> Set[str]:
    return _extract_the_all_variable(path, frozenset())


def extract_imported_api(path: Path) -> Set[str]:
    return extract_api(list_imports(path))


def fetch_api_for_path(path: Path) -> Set[str]:
    imported_api = extract_imported_api(path)
    symbols = extract_symbols(path)
    variables = extract_public_variables(path)
    the_all_variable = extract_the_all_variable(path)

    return imported_api | symbols | variables | the_all_variable


def fetch_api(paths: Set[Path]) -> dict:
    interface_paths = [Path(p / PACKAGE_INTERFACE) for p in paths]

    interfaces = [p for p in interface_paths if p.exists()]

    rows = [{i.parent.name: fetch_api_for_path(i)} for i in interfaces]

    return {k: v for row in rows for k, v in row.items()}



import ast
from collections.abc import Iterable
from functools import lru_cache
from pathlib import Path
from typing import List, Set, Union

typing_ns = "typing"
type_checking = "TYPE_CHECKING"


def parse_import(node: ast.Import) -> List[str]:
    return [name.name for name in node.names]


def extract_import_from(node: ast.ImportFrom) -> List:
    return (
        [f"{node.module}.{alias.name}" for alias in node.names]
        if node.names
        else [node.module]
    )


def parse_import_from(node: ast.ImportFrom) -> List[str]:
    return extract_import_from(node) if node.module and node.level == 0 else []


def parse_imports(node: ast.AST) -> List[str]:
    if isinstance(node, ast.Import):
        return parse_import(node)

    if isinstance(node, ast.ImportFrom):
        return parse_import_from(node)

    return []


def find_type_checking_body(node: ast.If) -> List[ast.stmt]:
    if isinstance(node.test, ast.Name) and node.test.id == type_checking:
        return node.body

    if isinstance(node.test, ast.Attribute) and node.test.attr == type_checking:
        if isinstance(node.test.value, ast.Name) and node.test.value.id == typing_ns:
            return node.body

    return []


def flatten(data: Iterable) -> list:
    return [item for nested in data for item in nested]


def parse_node(node: ast.AST) -> Union[dict, None]:
    if isinstance(node, ast.Import):
        return {"include": parse_import(node)}

    if isinstance(node, ast.ImportFrom):
        return {"include": parse_import_from(node)}

    if isinstance(node, ast.If):
        found = find_type_checking_body(node)
        parsed = flatten(parse_imports(f) for f in found)

        if not parsed:
            return None

        return {"exclude": parsed}

    return None


def parse_module(path: Path) -> ast.Module:
    with open(path.as_posix(), "r", encoding="utf-8", errors="ignore") as f:
        tree = ast.parse(f.read(), path.name)

    return tree


@lru_cache(maxsize=None)
def extract_imports(path: Path) -> List[str]:
    tree = parse_module(path)

    nodes = (parse_node(n) for n in ast.walk(tree))
    parsed_nodes = [n for n in nodes if n is not None]

    includes = [i for n in parsed_nodes for i in n.get("include", [])]
    excludes = {i for n in parsed_nodes for i in n.get("exclude", [])}

    return [i for i in includes if i not in excludes]


def extract_imports_and_flatten(py_modules: Iterable) -> Set[str]:
    return {i for m in py_modules for i in extract_imports(m)}


def is_python_file(path: Path) -> bool:
    return path.is_file() and path.suffix == ".py"


def find_files(path: Path) -> Iterable:
    return [path] if is_python_file(path) else path.rglob("*.py")


@lru_cache(maxsize=None)
def list_imports(path: Path) -> Set[str]:
    py_modules = find_files(path)

    return extract_imports_and_flatten(py_modules)


def fetch_all_imports(paths: Set[Path]) -> dict:
    rows = [{p.name: list_imports(p)} for p in paths]

    return {k: v for row in rows for k, v in row.items()}


def should_exclude(path: Path, excludes: Set[str]):
    return any(path.match(pattern) for pattern in excludes)


def list_excluded_imports(path: Path, excludes: Set[str]) -> Set[str]:
    py_modules = find_files(path)

    filtered = [p for p in py_modules if should_exclude(p, excludes)]

    return extract_imports_and_flatten(filtered)


def fetch_excluded_imports(paths: Set[Path], excludes: Set[str]) -> dict:
    rows = [{p.name: list_excluded_imports(p, excludes)} for p in paths]

    return {k: v for row in rows for k, v in row.items()}


def extract_top_ns_from_imports(imports: Set[str]) -> Set:
    return {imp.split(".")[0] for imp in imports}


def extract_top_ns(import_data: dict) -> dict:
    return {k: extract_top_ns_from_imports(v) for k, v in import_data.items()}



import ast
from functools import lru_cache
from pathlib import Path
from typing import FrozenSet, Set, Tuple, Union

from polylith.imports.parser import extract_imports, find_files, parse_module

WRAPPER_NODES = (ast.Await, ast.Expr, ast.NamedExpr, ast.Starred, ast.Subscript)
FN_NODES = (ast.FunctionDef, ast.AsyncFunctionDef, ast.Lambda)
SYMBOLS = (*FN_NODES, ast.ClassDef)


def extract_api_part(path: str) -> str:
    return path.rsplit(".", 1)[-1]


def extract_api(paths: Set[str]) -> Set[str]:
    return {extract_api_part(p) for p in paths}


def find_import_root_and_path(
    expr: ast.expr, parts: Tuple[str, ...] = ()
) -> Tuple[ast.expr, str]:
    """Builds a namespace when the expression is an Attribute or Name, otherwise empty."""
    if isinstance(expr, ast.Attribute):
        return find_import_root_and_path(expr.value, (*parts, expr.attr))

    namespace_parts = (*parts, expr.id) if isinstance(expr, ast.Name) else parts

    namespace = str.join(".", reversed(namespace_parts))

    return expr, namespace


def with_ns(usage: str, ns: str) -> str:
    return usage if str.startswith(usage, ns + ".") else f"{ns}.{usage}"


def match_usage(root_id: str, usage: str, entry: str) -> str:
    separator = "."

    part, *rest = str.split(usage, separator)

    if root_id == usage:
        return entry

    if root_id == part:
        return f"{separator}".join([entry, *rest])

    return usage


def find_matching_usage(expr: ast.expr, options: dict) -> Union[str, None]:
    ns = options["ns"]
    api_map = options["api_map"]
    allowed_prefixes = options["allowed_prefixes"]
    shadowed = options["shadowed"]

    root, usage = find_import_root_and_path(expr)

    if not isinstance(root, ast.Name):
        return None

    if root.id in shadowed:
        return None

    if root.id in api_map:
        entry = api_map[root.id]
        found = match_usage(root.id, usage, entry)

        return with_ns(found, ns)

    if any(usage.startswith(p + ".") for p in allowed_prefixes):
        return with_ns(usage, ns)

    return None


def parse_import_usage(node: ast.AST, options: dict) -> Union[str, None]:
    usage = None
    child = None

    if isinstance(node, ast.Attribute):
        usage = find_matching_usage(node, options)
        child = node.value
    elif isinstance(node, WRAPPER_NODES):
        child = node.value
    elif isinstance(node, ast.Call):
        usage = find_matching_usage(node.func, options)
        child = node.func
    elif isinstance(node, ast.UnaryOp):
        child = node.operand

    if usage:
        return usage

    return parse_import_usage(child, options) if child is not None else None


def collect_arg_names(
    fn: Union[ast.FunctionDef, ast.AsyncFunctionDef, ast.Lambda],
) -> Set[str]:
    args = fn.args

    names = {a.arg for a in args.posonlyargs + args.args + args.kwonlyargs}

    if args.vararg:
        names.add(args.vararg.arg)

    if args.kwarg:
        names.add(args.kwarg.arg)

    return names


def walk_usages(node: ast.AST, options: dict) -> Set[str]:
    if isinstance(node, FN_NODES):
        options = {
            **options,
            "shadowed": options["shadowed"] | frozenset(collect_arg_names(node)),
        }

    out = set()
    hit = parse_import_usage(node, options)

    if hit:
        out.add(hit)

    for child in ast.iter_child_nodes(node):
        out |= walk_usages(child, options)

    return out


def fetch_import_usages_in_module(path: Path, ns: str, imported: Set[str]) -> Set[str]:
    tree = parse_module(path)
    api_map = {extract_api_part(p): p for p in imported}

    options = {
        "ns": ns,
        "api_map": api_map,
        "allowed_prefixes": frozenset(api_map.values()),
        "shadowed": frozenset(),
    }
    return walk_usages(tree, options)


@lru_cache(maxsize=None)
def fetch_brick_import_usages(
    path: Path, ns: str, imported: FrozenSet[str]
) -> Set[str]:
    py_modules = find_files(path)

    found = {m: set(extract_imports(m)).intersection(imported) for m in py_modules}
    filtered = {k: v for k, v in found.items() if v}

    fetched = (fetch_import_usages_in_module(k, ns, v) for k, v in filtered.items())

    return {i for f in fetched if f for i in f}



from pathlib import Path
from typing import List, Set

from polylith import check, workspace


def get_brick_imports(
    root: Path, ns: str, bases: Set[str], components: Set[str]
) -> dict:
    bases_paths = workspace.paths.collect_bases_paths(root, ns, bases)
    comp_paths = workspace.paths.collect_components_paths(root, ns, components)

    brick_imports_in_bases = check.collect.extract_bricks(bases_paths, ns)
    brick_imports_in_components = check.collect.extract_bricks(comp_paths, ns)

    return {
        "bases": check.collect.with_unknown_components(
            root, ns, brick_imports_in_bases
        ),
        "components": check.collect.with_unknown_components(
            root, ns, brick_imports_in_components
        ),
    }


def without(key: str, bricks: Set[str]) -> Set[str]:
    return {b for b in bricks if b != key}


def sorted_usings(usings: Set[str], bases: Set[str], components: Set[str]) -> List[str]:
    usings_bases = sorted({b for b in usings if b in bases})
    usings_components = sorted({c for c in usings if c in components})

    return usings_components + usings_bases


def sorted_used_by(
    brick: str, bases: Set[str], components: Set[str], import_data: dict
) -> List[str]:
    brick_used_by = without(brick, {k for k, v in import_data.items() if brick in v})

    return sorted_usings(brick_used_by, bases, components)


def sorted_uses(
    brick: str, bases: Set[str], components: Set[str], import_data: dict
) -> List[str]:
    brick_uses = without(brick, import_data.get(brick, set()))

    return sorted_usings(brick_uses, bases, components)


def calculate_brick_deps(brick: str, bricks: dict, import_data: dict) -> dict:
    bases = bricks["bases"]
    components = bricks["components"]

    brick_used_by = sorted_used_by(brick, bases, components, import_data)
    brick_uses = sorted_uses(brick, bases, components, import_data)

    return {"used_by": brick_used_by, "uses": brick_uses}


def find_intersection_for_usings(usings: dict) -> Set[str]:
    uses = set(usings["uses"])
    used_by = set(usings["used_by"])

    return uses.intersection(used_by)


def find_bricks_with_circular_dependencies(bricks_deps: dict) -> dict:
    res = {k: find_intersection_for_usings(v) for k, v in bricks_deps.items()}

    return {k: v for k, v in sorted(res.items()) if v}



from pathlib import Path
from typing import List, Set

from polylith import bricks, deps, info, interface


def get_imports(root: Path, ns: str, bricks: dict) -> dict:
    bases = bricks["bases"]
    components = bricks["components"]
    brick_imports = deps.get_brick_imports(root, ns, bases, components)

    return {**brick_imports["bases"], **brick_imports["components"]}


def pick_name(data: List[dict]) -> Set[str]:
    return {b["name"] for b in data}


def get_bases(root: Path, ns: str, project_data: dict) -> Set[str]:
    if project_data:
        return set(project_data.get("bases", []))

    return pick_name(bricks.get_bases_data(root, ns))


def get_components(root: Path, ns: str, project_data: dict) -> Set[str]:
    if project_data:
        return set(project_data.get("components", []))

    return pick_name(bricks.get_components_data(root, ns))


def used_by_as_bricks(bricks: dict, brick_deps: dict) -> dict:
    bases = bricks["bases"]
    components = bricks["components"]

    used_by = brick_deps["used_by"]
    return {
        "bases": {b for b in used_by if b in bases},
        "components": {b for b in used_by if b in components},
    }


def run(root: Path, ns: str, options: dict):
    directory = options.get("directory")
    brick = options.get("brick")
    show_interface = options.get("show_interface")

    projects_data = info.get_projects_data(root, ns) if directory else []
    project = next((p for p in projects_data if directory in p["path"].as_posix()), {})

    bricks = {
        "bases": get_bases(root, ns, project),
        "components": get_components(root, ns, project),
    }

    imports = get_imports(root, ns, bricks)

    bricks_deps = {
        b: deps.calculate_brick_deps(b, bricks, imports)
        for b in set().union(*bricks.values())
    }

    circular_bricks = deps.find_bricks_with_circular_dependencies(bricks_deps)

    if brick and imports.get(brick):
        brick_deps = bricks_deps[brick]
        used_bricks = used_by_as_bricks(bricks, brick_deps)

        circular_deps = circular_bricks.get(brick)

        deps.print_brick_deps(brick, bricks, brick_deps, options)

        if show_interface:
            interface.report.print_brick_interface(root, ns, brick, used_bricks)

            interface.report.print_brick_interface_invalid_usage(
                root, ns, brick, used_bricks
            )

        if circular_deps:
            deps.print_brick_with_circular_deps(brick, circular_deps, bricks)

        return

    deps.print_deps(bricks, imports, options)
    deps.print_bricks_with_circular_deps(circular_bricks, bricks)



import difflib
from pathlib import Path
from typing import List, Set, Union

from polylith import configuration, info, repo, sync
from polylith.reporting import theme
from rich.console import Console
from rich.padding import Padding
from rich.prompt import Confirm, Prompt

console = Console(theme=theme.poly_theme)


def create_added_brick_message(bricks: Set[str], tag: str, project_name: str) -> str:
    number_of_bricks = len(bricks)
    plural = "s" if number_of_bricks > 1 else ""

    if tag == "base":
        grammar = f"base{plural}"
    else:
        grammar = f"component{plural}"

    return f"[data]Added {number_of_bricks} [{tag}]{grammar}[/] to the [proj]{project_name}[/] project.[/]"


def confirmation(diff: dict, project_name: str) -> None:
    pad = (1, 0, 0, 0)

    if not diff:
        nothing_added_message = f"[data]No bricks added to [proj]{project_name}[/][/]."
        console.print(Padding(nothing_added_message, pad))

        return

    bases = diff["bases"]
    components = diff["components"]

    bases_message = create_added_brick_message(bases, "base", project_name)
    console.print(Padding(bases_message, pad))

    if len(components) == 0:
        return

    components_message = create_added_brick_message(components, "comp", project_name)
    console.print(components_message)


def sort_bases_by_closest_match(bases: Set[str], name: str) -> List[str]:
    closest = difflib.get_close_matches(name, bases, cutoff=0.3)

    rest = sorted([b for b in bases if b not in closest])

    return closest + rest


def choose_base_for_project(
    root: Path,
    ns: str,
    project_name: str,
    possible_bases: Set[str],
) -> Union[str, None]:
    first, *_ = sort_bases_by_closest_match(possible_bases, project_name)

    if not Confirm.ask(
        prompt=f"[data]Do you want to add bricks to the [proj]{project_name}[/] project?[/]",
        console=console,
    ):
        return None

    question = "[data]What's the name of the Polylith [base]base[/] to add?[/]"

    base = Prompt.ask(
        prompt=question,
        console=console,
        default=first,
        show_default=True,
        case_sensitive=False,
    )

    all_bases = info.get_bases(root, ns)

    return next((b for b in all_bases if str.lower(b) == str.lower(base)), None)


def add_bricks_to_project(
    root: Path,
    ns: str,
    project_data: dict,
    possible_bases: Set[str],
) -> None:
    project_name = project_data["name"]
    found_base = choose_base_for_project(root, ns, project_name, possible_bases)

    if not found_base:
        confirmation({}, project_name)
        return

    diff = sync.calculate_needed_bricks(root, ns, project_data, found_base)

    sync.update_project(root, ns, diff)

    confirmation(diff, project_name)


def run(project_name: str) -> None:
    root = repo.get_workspace_root(Path.cwd())
    ns = configuration.get_namespace_from_config(root)

    possible_bases = info.find_unused_bases(root, ns)

    if not possible_bases:
        return

    projects_data = info.get_projects_data(root, ns)
    project_data = next((p for p in projects_data if p["name"] == project_name), None)

    if not project_data:
        return

    message = f"[data]Project [proj]{project_name}[/] created.[/]"
    console.print(Padding(message, (0, 0, 1, 0)))

    add_bricks_to_project(root, ns, project_data, possible_bases)



from functools import reduce
from itertools import zip_longest
from typing import List, Set, Tuple, Union

from polylith import output
from polylith.reporting import theme
from rich import box
from rich.console import Console
from rich.padding import Padding
from rich.table import Table


def to_col(brick: str, tag: str) -> str:
    name = "\n".join(brick)

    return f"[{tag}]{name}[/]"


def brick_status(bricks: Set[str], brick_name: str, imported: str) -> str:
    status = theme.check_emoji if imported in bricks and imported != brick_name else "-"

    return f"[data]{status}[/]"


def to_row(name: str, tag: str, brick_imports: dict, imported: List[str]) -> List[str]:
    bricks = brick_imports.get(name) or set()
    statuses = [brick_status(bricks, name, i) for i in imported]

    return [f"[{tag}]{name}[/]"] + statuses


def flatten_import(acc: Set[str], kv: Tuple[str, Set[str]]) -> set:
    key = kv[0]
    values = kv[1]

    return set().union(acc, values.difference({key}))


def flatten_imports(brick_imports: dict) -> Set[str]:
    """Flatten the dict into a set of imports, with the actual brick filtered away when existing as an import"""
    return reduce(flatten_import, brick_imports.items(), set())


def create_columns(
    imported_bases: List[str], imported_components: List[str]
) -> List[str]:
    base_cols = [to_col(brick, "base") for brick in imported_bases]
    comp_cols = [to_col(brick, "comp") for brick in imported_components]

    return comp_cols + base_cols


def create_rows(
    bases: Set[str], components: Set[str], import_data: dict, imported: List[str]
) -> List[List[str]]:
    base_rows = [to_row(b, "base", import_data, imported) for b in sorted(bases)]
    comp_rows = [to_row(c, "comp", import_data, imported) for c in sorted(components)]

    return comp_rows + base_rows


def find_longest_brick_name(bases: Set[str], components: Set[str]) -> str:
    bricks = set().union(bases, components)

    return max(bricks, key=len)


def calculate_brick_column_width(
    bases: Set[str], components: Set[str]
) -> Union[int, None]:
    longest = find_longest_brick_name(bases, components)

    return len(longest)


def print_deps(bricks: dict, import_data: dict, options: dict):
    bases = bricks["bases"]
    components = bricks["components"]
    flattened = flatten_imports(import_data)

    imported_bases = sorted({b for b in flattened if b in bases})
    imported_components = sorted({c for c in flattened if c in components})
    imported_bricks = imported_components + imported_bases
    save = options.get("save", False)

    table = Table(box=box.SIMPLE_HEAD)
    brick_width = calculate_brick_column_width(bases, components)
    table.add_column("[data]brick[/]", width=brick_width)

    cols = create_columns(imported_bases, imported_components)
    rows = create_rows(bases, components, import_data, imported_bricks)

    for col in cols:
        table.add_column(col, justify="center")

    for row in rows:
        table.add_row(*row)

    console = Console(theme=theme.poly_theme)
    console.print(table, overflow="ellipsis")

    if save:
        output.save(table, options, "deps")


def calculate_tag(brick: str, bases: Set[str]) -> str:
    return "base" if brick in bases else "comp"


def print_brick_deps(brick: str, bricks: dict, brick_deps: dict, options: dict):
    bases = bricks["bases"]

    save = options.get("save", False)

    brick_used_by = brick_deps["used_by"]
    brick_uses = brick_deps["uses"]

    tag = calculate_tag(brick, bases)

    console = Console(theme=theme.poly_theme)

    table = Table(box=box.SIMPLE_HEAD)
    table.add_column("[data]used by[/]")
    table.add_column(":backhand_index_pointing_left:")
    table.add_column(f"[{tag}]{brick}[/]")
    table.add_column(":backhand_index_pointing_right:")
    table.add_column("[data]uses[/]")

    for item in zip_longest(brick_used_by, brick_uses):
        used_by, uses = item

        used_by_tag = calculate_tag(used_by, bases) if used_by else ""
        uses_tag = calculate_tag(uses, bases) if uses else ""

        left = f"[{used_by_tag}]{used_by}[/]" if used_by else ""
        right = f"[{uses_tag}]{uses}[/]" if uses else ""

        row = [left, "", "", "", right]

        table.add_row(*row)

    console.print(table, overflow="ellipsis")

    if save:
        output.save(table, options, f"deps_{brick}")


def print_brick_with_circular_deps(brick: str, deps: Set[str], bricks: dict) -> None:
    bases = bricks["bases"]

    console = Console(theme=theme.poly_theme)

    tag = calculate_tag(brick, bases)

    with_tags = [f"[{calculate_tag(name, bases)}]{name}" for name in sorted(deps)]
    others = "[data],[/] ".join(with_tags)

    prefix = ":information:"
    message = f"[{tag}]{brick}[/] [data]is used by[/] {others} [data]and also uses[/] {others}[data].[/]"

    console.print(Padding(f"{prefix} {message}", (0, 0, 0, 1)), overflow="ellipsis")


def print_bricks_with_circular_deps(circular_bricks: dict, bricks: dict) -> None:
    for brick, deps in circular_bricks.items():
        print_brick_with_circular_deps(brick, deps, bricks)



from functools import partial
from pathlib import Path
from typing import Set

from polylith import check, configuration, deps, info


def _is_empty(brick: str, bricks_path: Path) -> bool:
    path = bricks_path / brick

    return not any(path.iterdir())


def _without_empty(
    root: Path, namespace: str, bricks: Set[str], brick_type: str
) -> Set[str]:
    theme = configuration.get_theme_from_config(root)

    if theme == "loose":
        bricks_path = root / f"{brick_type}/{namespace}"
    else:
        bricks_path = root / f"{brick_type}"

    return {brick for brick in bricks if not _is_empty(brick, bricks_path)}


def _calculate(root: Path, namespace: str, project_data: dict, bases: Set[str]) -> dict:
    components = set(project_data["components"])

    all_bases = info.get_bases(root, namespace)
    all_components = info.get_components(root, namespace)

    brick_imports = deps.get_brick_imports(root, namespace, bases, components)
    is_project = info.is_project(project_data)

    if is_project:
        brick_diff = check.collect.imports_diff(brick_imports, bases, components)
    else:
        all_bricks = set().union(all_bases, all_components)
        brick_diff = check.collect.diff(all_bricks, bases, components)

    bases_diff = {b for b in brick_diff if b in all_bases}
    comp_diff = {b for b in brick_diff if b in all_components}

    fn = partial(_without_empty, root, namespace)

    return {
        "name": project_data["name"],
        "path": project_data["path"],
        "is_project": is_project,
        "bases": bases_diff if is_project else fn(bases_diff, "bases"),
        "components": comp_diff if is_project else fn(comp_diff, "components"),
        "brick_imports": brick_imports,
    }


def calculate_diff(root: Path, namespace: str, project_data: dict) -> dict:
    bases = set(project_data["bases"])

    return _calculate(root, namespace, project_data, bases)


def calculate_needed_bricks(
    root: Path, namespace: str, project_data: dict, base: str
) -> dict:
    bases = {base}

    res = _calculate(root, namespace, project_data, bases)

    needed_bases = res["bases"].union(bases)

    return {**res, **{"bases": needed_bases}}



from functools import reduce
from io import StringIO
from pathlib import Path
from typing import Tuple, cast

from polylith.reporting import theme
from rich.console import Console
from rich.table import Table

replacements = {"\u2714": "X", "\U0001F448": "<-", "\U0001F449": "->"}


def replace_char(data: str, pair: Tuple[str, str]) -> str:
    return str.replace(data, *pair)


def adjust(data: str) -> str:
    return reduce(replace_char, replacements.items(), data)


def write_to_file(data: str, options: dict, command: str) -> None:
    adjusted = adjust(data)

    output = options["output"]
    fullpath = f"{output}/{command}.txt"

    Path(output).mkdir(parents=True, exist_ok=True)
    Path(fullpath).write_text(adjusted)


def save_recorded(console: Console, options: dict, command: str) -> None:
    exported = console.export_text()

    write_to_file(exported, options, command)


def save(table: Table, options: dict, command: str) -> None:
    console = Console(theme=theme.poly_theme, width=1024, file=StringIO())

    console.print(table, overflow="ellipsis")

    f = cast(StringIO, console.file)
    exported = f.getvalue()

    write_to_file(exported, options, command)



from pathlib import Path

from polylith.interface.collect import get_brick_interface
from polylith.interface.usage import check_brick_interface_usage, unified_usages
from polylith.reporting import theme
from rich.console import Console
from rich.table import Table


def print_brick_interface(root: Path, ns: str, brick: str, bricks: dict) -> None:
    brick_interface = get_brick_interface(root, ns, brick, bricks)

    if not brick_interface:
        return

    console = Console(theme=theme.poly_theme)

    tag = "base" if brick in bricks["bases"] else "comp"

    table = Table(box=None)

    message = f"[{tag}]{brick}[/] exposes:"
    table.add_column(message)

    for endpoint in sorted(brick_interface):
        *_ns, exposes = str.split(endpoint, ".")
        table.add_row(f"[data]{exposes}[/]")

    console.print(table, overflow="ellipsis")


def print_brick_interface_invalid_usage(
    root: Path, ns: str, brick: str, bricks: dict
) -> None:
    res = check_brick_interface_usage(root, ns, brick, bricks)

    invalid_usage = {
        brick: unified_usages(usages)
        for brick, usages in res.items()
        if not all(usages.values())
    }

    if not invalid_usage:
        return

    console = Console(theme=theme.poly_theme)

    table = Table(box=None)
    tag = "base" if brick in bricks["bases"] else "comp"

    for using_brick, usages in invalid_usage.items():
        using_tag = "base" if using_brick in bricks["bases"] else "comp"

        for using in usages:
            used = str.replace(using, f"{ns}.{brick}.", "")
            prefix = f"Found in [{using_tag}]{using_brick}[/]"
            middle = f"[data]{used}[/] is not part of the public interface of [{tag}]{brick}[/]"

            message = f":information: {prefix}: {middle}."

            table.add_row(f"{message}")

    console.print(table, overflow="ellipsis")



from pathlib import Path

from cleo.helpers import option
from poetry.console.commands.command import Command
from polylith import commands, configuration, repo
from polylith.poetry.commands import command_options


class DepsCommand(Command):
    name = "poly deps"
    description = "Visualize the dependencies between <comment>bricks</>."

    options = [
        option(
            long_name="brick",
            description="Shows dependencies for selected brick",
            flag=False,
        ),
        option(
            long_name="interface",
            description="Show the brick interface",
            flag=True,
        ),
        command_options.save,
    ]

    def handle(self) -> int:
        directory = self.option("directory")
        brick = self.option("brick")
        save = self.option("save")
        interface = self.option("interface")

        root = repo.get_workspace_root(Path.cwd())
        ns = configuration.get_namespace_from_config(root)

        dir_path = Path(directory).as_posix() if directory else None
        output = configuration.get_output_dir(root, "deps") if save else None

        options = {
            "directory": dir_path,
            "brick": brick,
            "save": save,
            "output": output,
            "show_interface": interface,
        }
        commands.deps.run(root, ns, options)

        return 0



from pathlib import Path
from typing import Union

from polylith import info, interactive, sync


def is_project_without_bricks(project_data: dict) -> bool:
    bases = project_data["bases"]
    components = project_data["components"]

    return not bases and not components


def choose_base(root: Path, ns: str, project_data: dict) -> Union[str, None]:
    possible_bases = info.find_unused_bases(root, ns)

    if not possible_bases:
        return None

    return interactive.project.choose_base_for_project(
        root, ns, project_data["name"], possible_bases
    )


def can_run_interactive_mode(project_data: dict, options: dict) -> bool:
    is_quiet = options["quiet"]

    if is_quiet:
        return False

    return info.is_project(project_data) and is_project_without_bricks(project_data)


def calculate_brick_diff(
    root: Path, ns: str, project_data: dict, options: dict
) -> dict:
    if can_run_interactive_mode(project_data, options):
        base = choose_base(root, ns, project_data)

        if base:
            return sync.calculate_needed_bricks(root, ns, project_data, base)

    return sync.calculate_diff(root, ns, project_data)


def run(root: Path, ns: str, project_data: dict, options: dict):
    is_quiet = options["quiet"]
    is_verbose = options["verbose"]

    diff = calculate_brick_diff(root, ns, project_data, options)

    sync.update_project(root, ns, diff)

    if is_quiet:
        return

    sync.report.print_summary(diff)

    if is_verbose:
        sync.report.print_brick_imports(diff)



from polylith.sync import report
from polylith.sync.collect import calculate_diff, calculate_needed_bricks
from polylith.sync.update import update_project

__all__ = ["report", "calculate_diff", "calculate_needed_bricks", "update_project"]



from poetry.console.commands.command import Command
from polylith.commands.create import create


def try_create(command: Command, fn) -> int:
    name = command.option("name")
    description = command.option("description")

    try:
        create(name, description, fn)
    except ValueError as e:
        command.line_error(f"Did not create: <error>{e}</error>")

        return 1

    return 0



from pathlib import Path
from typing import Dict, FrozenSet, Set

from polylith import imports
from polylith.interface.collect import get_brick_imports, get_brick_interface
from polylith.workspace.paths import collect_bases_paths, collect_components_paths


def extract_endpoint(usage: str) -> str:
    separator = "."

    if separator not in usage:
        return usage

    _ns, _brick, *rest = str.split(usage, separator)

    return rest[0]


def unified_usages(usages: dict) -> Set[str]:
    filtered = {k for k, v in usages.items() if not v}

    return {extract_endpoint(f) for f in filtered}


def to_imported_api(brick_imports: Set[str]) -> Set[str]:
    return {imports.usages.extract_api_part(b) for b in brick_imports}


def equals_or_starts_with(brick_import: str, brick: str, ns: str) -> bool:
    brick_ns = f"{ns}.{brick}"
    brick_ns_part = f"{brick_ns}."

    return brick_import == brick_ns or str.startswith(brick_import, brick_ns_part)


def filter_by_brick(brick_imports: Set[str], brick: str, ns: str) -> Set[str]:
    return {b for b in brick_imports if equals_or_starts_with(b, brick, ns)}


def is_within_namespace(current: str, namespaces: Set[str]) -> bool:
    return any(current.startswith(i) for i in namespaces)


def check_usage(usings: Set[str], brick_interface: Set[str]) -> dict:
    return {u: is_within_namespace(u, brick_interface) for u in usings}


def frozen(data: Dict[str, Set[str]], key: str) -> FrozenSet[str]:
    return frozenset(data.get(key) or set())


def check_brick_interface_usage(root: Path, ns: str, brick: str, bricks: dict) -> dict:
    brick_interface = get_brick_interface(root, ns, brick, bricks)
    bases = bricks["bases"]
    components = bricks["components"]

    brick_imports = get_brick_imports(root, ns, bases, components)
    by_brick = {k: filter_by_brick(v, brick, ns) for k, v in brick_imports.items()}

    bases_paths = collect_bases_paths(root, ns, bases)
    comp_paths = collect_components_paths(root, ns, components)
    paths = bases_paths.union(comp_paths)

    usage = {
        p.name: imports.fetch_brick_import_usages(p, ns, frozen(by_brick, p.name))
        for p in paths
    }

    checked = {k: check_usage(v, brick_interface) for k, v in usage.items()}

    return checked



from pathlib import Path

from polylith import configuration, info
from polylith.dirs import create_dir
from polylith.files import create_file
from polylith.interface import create_interface
from polylith.readme import create_brick_readme


def validate(root: Path, namespace: str, name: str) -> None:
    bases = info.get_bases(root, namespace)
    components = info.get_components(root, namespace)

    if name in bases:
        message = f"{name} already exists in bases."
        raise ValueError(message)

    if name in components:
        message = f"{name} already exists in components."
        raise ValueError(message)


def create_brick(root: Path, options: dict) -> None:
    modulename = options["modulename"]
    package = options["package"]
    namespace = options["namespace"]

    validate(root, namespace, package)

    path_kwargs = {
        k: v for k, v in options.items() if k in {"brick", "namespace", "package"}
    }

    brick_structure = configuration.get_brick_structure_from_config(root)
    resources_structure = configuration.get_resources_structure_from_config(root)

    brick_path = brick_structure.format(**path_kwargs)
    resources_path = resources_structure.format(**path_kwargs)

    d = create_dir(root, brick_path)
    create_file(d, f"{modulename}.py")
    create_interface(d, options)

    if configuration.is_readme_generation_enabled(root):
        create_brick_readme(root / resources_path, options)



from pathlib import Path

from polylith import imports
from polylith.interface.parser import fetch_api
from polylith.workspace.paths import collect_bases_paths, collect_components_paths


def get_brick_interface(root: Path, ns: str, brick: str, bricks: dict) -> set:
    bases = bricks["bases"]
    paths = {brick}

    fn = collect_bases_paths if brick in bases else collect_components_paths

    brick_paths = fn(root, ns, paths)
    bricks_api = fetch_api(brick_paths)
    brick_api = bricks_api.get(brick) or set()
    brick_ns = f"{ns}.{brick}"

    return {f"{brick_ns}.{endpoint}" for endpoint in brick_api}


def get_brick_imports(root: Path, ns: str, bases: set, components: set) -> dict:
    bases_paths = collect_bases_paths(root, ns, bases)
    components_paths = collect_components_paths(root, ns, components)

    in_bases = imports.fetch_all_imports(bases_paths)
    in_comps = imports.fetch_all_imports(components_paths)

    extracted_bases = imports.extract_brick_imports_with_namespaces(in_bases, ns)
    extracted_components = imports.extract_brick_imports_with_namespaces(in_comps, ns)

    return {**extracted_bases, **extracted_components}



from cleo.helpers import option

alias = option(
    long_name="alias",
    description="alias for a third-party library, useful when an import differ from the library name",
    flag=False,
    multiple=True,
)


save = option(
    long_name="save",
    description="Store the contents of this command to file",
    flag=True,
)


short = option(
    long_name="short",
    short_name="s",
    description="Print short view",
    flag=True,
)

since = option(
    long_name="since",
    description="Changed since a specific tag",
    flag=False,
)

strict = option(
    long_name="strict",
    description="More strict checks when matching name and version of third-party libraries and imports.",
    flag=True,
)



from pathlib import Path

from poetry.console.commands.command import Command
from polylith import commands, configuration, repo
from polylith.poetry.commands import command_options


class InfoCommand(Command):
    name = "poly info"
    description = "Info about the <comment>Polylith</> workspace."

    options = [command_options.save, command_options.short]

    def handle(self) -> int:
        short = True if self.option("short") else False
        save = self.option("save")

        root = repo.get_workspace_root(Path.cwd())
        output = configuration.get_output_dir(root, "info") if save else None

        options = {
            "short": short,
            "save": save,
            "output": output,
        }
        commands.info.run(root, options)

        return 0



from polylith.deps.core import (
    calculate_brick_deps,
    find_bricks_with_circular_dependencies,
    get_brick_imports,
)
from polylith.deps.report import (
    print_brick_deps,
    print_brick_with_circular_deps,
    print_bricks_with_circular_deps,
    print_deps,
)

__all__ = [
    "calculate_brick_deps",
    "find_bricks_with_circular_dependencies",
    "get_brick_imports",
    "print_brick_deps",
    "print_brick_with_circular_deps",
    "print_bricks_with_circular_deps",
    "print_deps",
]
